#include <vector>
#include <cassert>

#include "graph.h"
#include "flow.h"
#include "connectivity.h"

#define F_IN(a) a->fvert[0]
#define F_OUT(a) a->fvert[1]

static const long infty = 100000000;

Edge *fake_edge = (Edge *) 10;

void construct_flow_graph(Graph &G, F_Graph &Gf) {
	for (Vertex *a: G.V) {
		// create a "a_in" and "a_out" vertex connected by a single edge
		a->fvert.push_back(create_vertex(Gf));
		a->fvert.push_back(create_vertex(Gf));
		create_edge(Gf, F_IN(a), F_OUT(a), a->weight, fake_edge);
	}

	for (Edge *e: G.E) {
		// for each edge {a, b} create the edges (a, b) and (b, a) with ininite capacity
		create_edge(Gf, F_OUT(e->end[0]), F_IN(e->end[1]), infty, e);
		create_edge(Gf, F_OUT(e->end[1]), F_IN(e->end[0]), infty, e);
	}

}

void reset_flow_graph(F_Graph &Gf) {
	for (F_Edge *e: Gf.E) {
		if (e->a == nullptr) {
			assert(e->b == nullptr);
			continue;
		}

		long cap = e->cap + e->rev->cap;
		if (e->dual != nullptr && e->rev->dual == nullptr) {
			e->cap = cap;
			e->rev->cap = 0;
		}
		else assert(e->dual == nullptr && e->rev->dual != nullptr);
	}
}

long vertex_connectivity_helper(F_Graph &Gf, F_Vertex *s, F_Vertex *t, Vertex *u, Vertex *v) {
	/* adapt generic flow network */
	assert(s->edges.size() == 0);
	assert(t->edges.size() == 0);

	F_Edge *e1 = create_edge(Gf, s, F_OUT(v), infty, fake_edge);
	F_Edge *e2 = create_edge(Gf, F_IN(u), t, infty, fake_edge);
	
	assert(s->edges.size() == 1);
	assert(t->edges.size() == 1);
	assert(s->edges.front().second->cap == infty && s->edges.front().second->rev->cap == 0);
	assert(t->edges.front().second->cap == 0 && t->edges.front().second->rev->cap == infty);

	long flow = compute_max_flow(Gf, s, t);

	/* restore generic flow network */
	reset_flow_graph(Gf);
	for (auto edge: s->edges)
		assert(edge.second->cap == infty);
	for (auto edge: t->edges)
		assert(edge.second->rev->cap == infty);

	F_Vertex *tmp = F_OUT(v);
	assert(tmp->edges[tmp->edges.size() - 1].first == s);
	tmp->edges.resize(tmp->edges.size() - 1);

	tmp = F_IN(u);
	assert(tmp->edges[tmp->edges.size() - 1].first == t);
	tmp->edges.resize(tmp->edges.size() - 1);

	s->edges.clear();
	t->edges.clear();

	/* dirty cleanup for the edges */
	// edges will be freed by the destructor of Gf
	e1->a = e1->b = nullptr;
	e2->a = e2->b = nullptr;

	return flow;
}

/* computes weighted connectivity */
// based on algorithm 11 from https://www.cse.msu.edu/~esfahani/book_chapter/Graph_connectivity_chapter.pdf 
long vertex_connectivity(Graph &G) {
	if (G.V.size() == 0)
		return 0;

	/* initial connectivity is n-1 (compute general case with weights) */
	long connectivity = 0;
	long long min_weight = infty;
	for (Vertex *v: G.V) {
		connectivity += v->weight;
		min_weight = min(min_weight, v->weight);
	}

	connectivity -= min_weight;

	/* find minimum degree vertex */
	Vertex *v = G.V.front();
	for (Vertex *a: G.V) {
		if (a->deg < v->deg)
			v = a;
	}

	/* setup generic flow graph */
	F_Graph Gf;
	construct_flow_graph(G, Gf);

	F_Vertex *s = create_vertex(Gf);
	F_Vertex *t = create_vertex(Gf);

	/* find minumum vertex cut that does not remove v */
	for (Vertex *u: G.V) {
		if (u == v)
			continue;
		/* check that u and v are not adjacent */
		bool adjacent = false;
		for (auto edge: v->edges) {
			if (edge.first == u) {
				adjacent = true;
				break;
			}
		}
		if (adjacent)
			continue;

		long flow = vertex_connectivity_helper(Gf, s, t, u, v);
		connectivity = min(flow, connectivity);
		if (connectivity == 0)
			goto done;
	}

	/* find minumum vertex cut that separates some neighbours of v */
	for (auto edge: v->edges) {
		Vertex *a = edge.first;

		for (auto edge2: v->edges) {
			Vertex *b = edge2.first;
			if (a->id >= b->id)
				continue;

			/* check that a and b are not adjacent */
			bool adjacent = false;
			for (auto edge: a->edges) {
				if (edge.first == b) {
					adjacent = true;
					break;
				}
			}
			if (adjacent)
				continue;

			long flow = vertex_connectivity_helper(Gf, s, t, a, b);
			connectivity = min(flow, connectivity);
			if (connectivity == 0)
				goto done;
		}
	}
done:

	for (Vertex *a: G.V) {
		a->fvert.clear();
	}

	return connectivity;
}
