#include <vector>
#include <cassert>

#include "graph.h"
#include "snapshot.h"
#include "connectivity.h"

static void dfs(Vertex *v, size_t nr, vector<Vertex *> &component) {
	if (v->component != 0) {
		assert(v->component == nr);
		return;
	}
	v->component = nr;
	component.push_back(v);

	for (auto edge: v->edges) {
		Vertex *a = edge.first;
		dfs(a, nr, component);
	}
}

static bool induced_dfs(Vertex *v, vector<Vertex *> &P)
{
	if (v->visited)
		return false;
	// check that adding v does not break the induced path
	size_t count = 0;
	for (auto edge: v->edges)  {
		if (edge.first->visited)
			count++;
	}

	if (P.size() != 0 && count != 1)
		return false;

	bool found = false;
	P.push_back(v);
	v->visited = true;

	if (P.size() == 4) {
		if (P[3]->marked)
			found = true;
		else	found = false;
	}
	else {
		for (auto edge: v->edges)  {
			Vertex *a = edge.first;

			if (induced_dfs(a, P)) {
				found = true;
				break;
			}
		}
	}

	P.pop_back();
	v->visited = false;
	return found;
}

// check if there is an induced P4 that starts in s and ends in some vertex t in T
static bool induced_p4_exists(Vertex *s, vector<Vertex *> T) {
	vector<Vertex *> X;
	vector<Vertex *> N;

	// mark vertices in T
	for (Vertex *t: T) {
		assert(!t->marked2);
		t->marked2 = true;
	}

	// mark all neighbours of s
	for (auto edge: s->edges) {
		Vertex *a = edge.first;
		assert(!a->marked);
		a->marked = true;
	}

	// filter out nodes adjacent to s
	N.clear();
	for (Vertex *t: T) {
		if (!t->marked) {
			N.push_back(t);
		}
	}

	// get all vertices with dist 1 to any vertex in N
	X.clear();
	for (Vertex *n: N) {
		for (auto edge: n->edges) {
			Vertex *x = edge.first;
			if (!x->visited) {
				x->visited = true;
				X.push_back(x);
			}
		}	
	}
	for (Vertex *x: X) {
		assert(x->visited);
		x->visited = false;
	}

	// filter out nodes adjacent to s
	N.clear();
	for (Vertex *x: X) {
		if (!x->marked) {
			N.push_back(x);
		}
	}

	// magic step
	bool exists = false;
	X.clear();
	for (Vertex *n: N) {
		size_t count = 0;
		for (auto edge: n->edges) {
			Vertex *y = edge.first;
			// if y in T
			if (y->marked2) {
				y->marked3 = true;
				count++;
			}
		}	

		for (auto edge: n->edges) {
			Vertex *x = edge.first;
			if (!x->marked)
				continue;

			size_t bad = 0;

			for (auto edge2: x->edges) {
				Vertex *y = edge2.first;
				if (y->marked2 && y->marked3)
					bad++;
			}

			if (bad < count) {
				if (x->marked) {
					exists = true;
					break;
				}
			}
		}	

		for (auto edge: n->edges) {
			Vertex *y = edge.first;
			// if y in T
			if (y->marked2) {
				y->marked3 = false;
			}
		}

		if (exists)
			break;

	}

	// cleanup
	for (auto edge: s->edges) {
		Vertex *a = edge.first;
		assert(a->marked);
		a->marked = false;
	}
	for (Vertex *t: T) {
		assert(t->marked2);
		t->marked2 = false;
	}

	return exists;
}

static bool is_bridge(vector<Vertex *> neighbours) {
	// find induced P4 stuv such that s,v is in neighbours
	bool bridge = false;

	for (Vertex *a: neighbours) {
		assert(!a->marked);
		a->marked = true;
	}

	for (Vertex *s: neighbours) {
		vector<Vertex *> P;
		s->marked = false;

		assert(!s->visited);
		bool found = induced_dfs(s, P);

		if (found) {
			bridge = true;
			break;
		}

		s->marked = true;
	}

	for (Vertex *a: neighbours) {
		a->marked = false;
	}

	return bridge;
}

static bool safety_guarantee(vector<Vertex *> neighbours, const long long k) {
	// check that strength(a, b) >= k for each a,b
	for (Vertex *a: neighbours) {
		// unsafe vertices
		vector<Vertex *> U;

		for (auto edge: a->edges) {
			Vertex *b = edge.first;
			assert(!b->marked);
			b->marked = true;
		}

		// find unsafe vertices
		for (Vertex *b: neighbours) {
			if (a == b)
				continue;
			if (b->marked)
				continue;

			long long strength = 0;
#if 1
			for (auto edge: b->edges) {
				Vertex *c = edge.first;
				if (c->marked) {
					strength += c->weight;
					if (c->permanent)
						goto skip;	// strength = infinity
				}
			}
#endif

			if (strength < k) {
				U.push_back(b);
			}
		skip:
			;
		}

		for (auto edge: a->edges) {
			Vertex *b = edge.first;
			b->marked = false;
		}

		// see if there is an induced P4 starting in a and ending in U
		if (induced_p4_exists(a, U))
			return false;
	}

	return true;
}

static bool is_2club(vector<Vertex *> component) {
	vector<Vertex *> marked;
	vector<Vertex *> current;
	vector<Vertex *> next;
	int dist;

	for (Vertex *a: component) {
		// find restricted P4 beginning with a
		bool found = false;

		dist = 1;
		marked.push_back(a);
		assert(!a->marked);
		a->marked = true;

		for (auto edge: a->edges) {
			Vertex *b = edge.first;
			b->marked = true;
			b->tmp    = a;

			current.push_back(b);
			marked.push_back(b);
		}


		// BFS
		while (dist != 3) {
			for (Vertex *b: current) {
				for (auto edge: b->edges) {
					Vertex *c = edge.first;
					if (c->marked)
						continue;

					c->marked = true;
					c->tmp    = b;

					next.push_back(c);
					marked.push_back(c);
				
				}
			}

			current.clear();
			swap(current, next);
			dist++;
		}

		if (current.size() != 0) {
			found = true;
		}

		// cleanup
		for (Vertex *b: marked) {
			assert(b->marked);
			b->marked = false;
			b->tmp = nullptr;
		}
		marked.clear();
		current.clear();
		next.clear();

		if (found)
			return false;
	}

	return true;
}

bool tc_permanent_club_single(Graph &G, long long &k, Vertex *v) {
	vector<Vertex *> neighbours;
	bool can_delete_v = false;

	if (v->deg == 0)
		return false;
	if (v->weight != 1)
		return false;
	if (v->permanent)
		return false;

	struct graph_snapshot snapshot = create_snapshot(G);

	for (Vertex *a: G.V) {
		assert(!a->marked);
		a->component = 0;
	}

	for (auto edge: v->edges) {
		neighbours.push_back(edge.first);
	}

	// see if v is a cut vertex
	delete_vertex(G, v);

	bool has_2club = false;
	size_t num_components = 0;
	vector< pair< vector<Vertex *>, bool > > components;
	
	for (Vertex *a: neighbours) {
		if (a->component != 0)
			continue;
		num_components++;
		vector<Vertex *> component;
		assert(component.size() == 0);
		// find component containing a
		dfs(a, num_components, component);

		bool two_club = is_2club(component);
		components.push_back(make_pair(component, two_club));
		if (two_club) {
			has_2club = true;
		}
	}

	if (!has_2club)
		goto cleanup;

	for (size_t i = 0; i < components.size(); i++) {
		if (components[i].second == true)
			continue;

		vector<Vertex *> N;
		for (Vertex *a: neighbours) {
			if (a->component == i+1)
				N.push_back(a);
		}

		if (!safety_guarantee(N, k)) {
			goto cleanup;
		}
	}

	// undo removing v
	restore_snapshot(G, snapshot);
	assert(v->deg != 0);

	// apply rule
	for (size_t i = 0; i < components.size(); i++) {
		if (components[i].second == false)
			continue;
		// mark vertices as permanent
		for (Vertex *a: components[i].first)
			mark_permanent(G, a);

		/* shrink this permanent 2-club down to a path */
		vector<Vertex *> marked;
		vector<Vertex *> current;
		vector<Vertex *> next;

		v->marked = true;
		marked.push_back(v);
		current.push_back(v);

		// BFS
		int dist;
		for (dist = 1; dist <= 4; dist++) {
			for (Vertex *b: current) {
				for (auto edge: b->edges) {
					Vertex *c = edge.first;
					if (c->marked)
						continue;
					// make sure we're BFSing only in the 2-club and not outside of it
					if (c->component != i+1) {
						assert(dist == 1);
						continue;
					}

					c->marked = true;
					c->tmp    = b;

					next.push_back(c);
					marked.push_back(c);
				
				}
			}

			if (next.size() == 0) {
				break;
			}
			assert(dist != 4); // we can't have a vertex with distance 4 to v

			current.clear();
			swap(current, next);
		}

		dist = dist - 1;
		// current contains vertices with maximum distance = dist < 4 to v
		assert(current.size() != 0);

		for (Vertex *b: marked)
			b->marked = false;

		/* mark the path we just found */
		Vertex *p = current[0];	
		while (true) {
			if (p == v)
				break;
			p->marked = true;
			p = p->tmp;
		}

		// delete all vertices in this component except the marked vertices
		for (Vertex *b: components[i].first) {
			b->tmp = nullptr;
			if (!b->marked)
				delete_vertex(G, b);
			else	b->marked = false;
		}

		if (dist == 3) {
			// the path we found forms a restricted P4 together with v
			can_delete_v = true;
		}

	}

	if (can_delete_v) {
		add_to_vds(G, k, v);

		#if 1
		for (size_t i = 0; i < components.size(); i++) {
			if (components[i].second == false)
				continue;
			for (Vertex *b: components[i].first) {
				if (b->deg != 0)
					delete_vertex(G, b);
			}
		}
		#endif
	}

	return true;

cleanup:
	restore_snapshot(G, snapshot);
	assert(v->deg != 0);
	return false;
}

bool tc_permanent_club(Graph &G, long long &k) {
	bool success = false;

	// TODO: is this faster with or without
	//if (vertex_connectivity(G) > 1)
	//	return false;

	vector<Vertex *> V_copy(G.V);
	for (Vertex *v: V_copy) {
		if (v->deg == 0)
			continue;

		if (tc_permanent_club_single(G, k, v))
			success = true;
	}

	return success;
}
